#ifndef _FUSION_CLUABINDVM_HPP_INCLUDED
#define _FUSION_CLUABINDVM_HPP_INCLUDED




#include "../fusCompileConfig.hpp"
#include "IScriptInterpreter.hpp"

#include <luabind/lua_include.hpp>
#include <luabind/luabind.hpp>
#include <string>

namespace fus {

//std::string const&,
typedef void (*TOnLuaErrorCallback)(std::string const&);



extern TOnLuaErrorCallback onLuaErrorFunction;

// Utilities
int onLuaError(lua_State*);
int onLuaPanic(lua_State *s);
int onNewIndex(lua_State *s);
int onIndex(lua_State *L);
void stackDump (lua_State *L);

//lua_State *L
//int specialNewIndex();

    //typedef boost::function<void,const std::string&> f;

// Pour passer en mode debug !!
    //int lua_sethook (lua_State *L, lua_Hook f, int mask, int count);


/** \brief TODO c
 * \todo check that string won't throw any exceptio ntoo
 * \param
 * \param
 * \return
 *
 */

class CLuaException : public std::exception
{
    public:
        //!< Constructor
        CLuaException(const std::string& message);

        /// Destructor ( throw() shows destructor won't throw any exception )
        virtual ~CLuaException() throw();

        /// Return message associated to this exception
        virtual const char* what() const throw();

    protected :
    std::string mMessage;       //!< Message associé à l'exception

};


/** \brief Set of helper functions
 *
 * \param
 * \param
 * \return
 *
 */

class CLuabindVM :
    public CLuaVM
{

public:

    CLuabindVM(lua_State* = 0);
    virtual ~CLuabindVM();


    static lua_State* createVM();
    static void loadVM( lua_State*);


    //static TOnLuaErrorCallback registerErrorHandlers(TOnLuaErrorCallback&);
    static void registerErrorHandlers(TOnLuaErrorCallback&);

    /**
    Retreive string on top of stack
    **/
    bool retrieveStringFromStack(std::string&);

    void reset();

    /** \brief
     *
     * \param
     * \param
     * \return true if error
     *
     */
    //static bool checkReturnValue(const int&);
    static bool handleError(lua_State* state,std::string&);
    const std::string& getErrorDescription() {return _errorMessage; };
    void setErrorDescription(const int&);
    bool bind(const luabind::scope& scope);
    bool RunCommand(const char* param);
    bool RunFile(const char* param) ;
    bool RunChunk(const int& array,const int& index) ;


//    bool loadFile(const char*) {
//        int temp;
//        return loadFile();
//    }

    /** \brief
     *
     * \param Filename
     * \param Mode Might be "b","t" or "bt" depending on if you want to load from binary or text
     * \return
     *
     */
    bool loadFile(const char*,const char* mode);


    // Weird ?
    bool loadChunkFromGlobals(const int&);

    int saveTopChunkIntoGlobals();

    void getGlobals( luabind::object& );
    lua_State* getVM() { return _vm;};
    virtual bool execute();

    // Prevents creation of new attributes to a class
    bool setSpecialNewIndex(const char*);


protected:
//    static TOnLuaErrorCallback onLuaErrorFunction;


    void destroyVM();

    //int loadString(const char*);

    lua_State* _vm;
    std::string _errorMessage;

};



}

#endif
