/**
* \file options.cpp
*
*/

#include "COptionsState.hpp"
#include "CMenuState.hpp"
#include "../game_engine.hpp"
#include "../config.hpp"
#include <vector>
#include <algorithm>
#include "../utils/CBitSet.hpp"
#include "CErrorState.hpp"
#include <fusion/core/CStringBuilder.hpp>
//#include <fusion/macros.hpp>
#include <xeffects/XEffects.h>
#include "../events/CBindCollectionPlayer.hpp"
#include "../events/CBindDescriptor.hpp"
#include "../events/CBindFromEventGenerator.hpp"

using namespace irr;
using namespace core;
using namespace video;
using namespace scene;
using namespace gui;
//using namespace input;


DECLARE_GAME_SINGLETON( COptionState )


#define GET_MAPPING(numero) (GameConfig.MappingPlayer1)
//engine().GameConfig.
#define GENERATE_DESCRIPTION(keymap)  keymap.generateDescription().c_str()

//template<int id,int numero> struct TRANSLATED_ID() { return (id*1000 + 10000 + numero);};
// Numero should be either 0 or 1
int TRANSLATED_ID(const int& id, const int& numero) { return (id*1000 + 10000 + numero);};


//::EId
std::pair<NPlayerInput::EId,int>
UNTRANSLATE_ID(const int& id)
{
    //button1->setText( GET_MAPPING(1)[std::make_pair(id,0)].generateDescription().c_str() );
    int iId = static_cast<int>(id);
    int numero = iId%2;

    //::EId
    return std::make_pair( static_cast<NPlayerInput::EId>( (iId - numero -10000)/1000 ) , numero );
}

/*
inline std::pair<NPlayerInput::EId,std::size_t>
UNTRANSLATE_ID(const int& translatedId) {

    std::pair<NPlayerInput::EId,std::size_t> ret;

    // Keep just units
    ret.second = translatedId%1000;
    ret.first = static_cast<NPlayerInput::EId>( (translatedId - 10000 - ret.second)/1000 );
    //if(translatedId%1000 != 0){
    return ret;
    //}
}
*/


COptionState::COptionState( SIngameConfig const& config) :
//_pos(pos),
IGameState("Options menu",true,true,true,false,true)
, _buttonToUpdate(0)
, _overallTab (0)
//, Background( background )
, GameConfig( config )
{
//    const irr::core::rectf& pos

}





COptionState::~COptionState()
{
    _INFO << "OPTIONS DESTRUCTOR";
//    _overallTab->remove();
}






void
COptionState::setButtonValue( NPlayerInput::EId const& id)
{


    irr::gui::IGUIButton* button1 = 0, *button2 = 0;

    button1 = static_cast<irr::gui::IGUIButton*>( _tabControl->getElementFromId( TRANSLATED_ID(id,0) ,true) );
    button2 = static_cast<irr::gui::IGUIButton*>( _tabControl->getElementFromId( TRANSLATED_ID(id,1) ,true) );

    std::wstring desc;

    GET_MAPPING(1).getBind(std::make_pair(id,0)).generateDescription( desc );

    button1->setText( desc.c_str() );
    desc.clear();

    GET_MAPPING(1).getBind(std::make_pair(id,1) ).generateDescription(desc);
    button2->setText( desc.c_str()  );

}




// engine->driver
void
COptionState::setToConfigValues()
{
    _LOG_DEBUG << "GUI set to match saved config values";

    //selectCurrentVideoMode
    IVideoModeList* list = engine().device()->getVideoModeList();

    //stringw currentMode = videoModeToText(driver->);
    /////////////////////////////////////////////////////////////////
    ///////     select current resolution
    /////////////////////////////////////////////////////////////////
    for(s32 i = 0; i < list->getVideoModeCount(); i += 1)
    {

        //compares 2 core::dimension2d< u32 >
        //driver->
        if((list->getVideoModeResolution(i) ==  engine().getScreenSize ()) )
        {
            _comboBox->setSelected(i);
        }

    }

// -- Graphic Tab
    /////////////////////////////////////////////////////////////////
    ///////     select current resolution
    /////////////////////////////////////////////////////////////////
//    for(std::size_t i = 0; i < NPlayerInput::Count; ++i)
//    {
//        // get specific mapping
//        CBindDescriptor& key = engine().GameConfig.Keymap1[i];
//
//        irr::gui::IGUIButton* button = static_cast<irr::gui::IGUIButton*>(_tabControl->getElementFromId( (NGuiId::Count + i) ,true));
//        button->setText( key.generateDescription().c_str());
//        button->setToolTipText(L"Click to change mapping");
//    }

    /////////////////////////////////////////////////////////////////
    ///////     set Control buttons text
    /////////////////////////////////////////////////////////////////
    setButtonValue( NPlayerInput::MoveForward );
    setButtonValue( NPlayerInput::MoveBackward );

    setButtonValue( NPlayerInput::MoveRight );
    setButtonValue( NPlayerInput::MoveLeft );
    setButtonValue( NPlayerInput::ModeUp );
    setButtonValue( NPlayerInput::ModeDown );
    setButtonValue( NPlayerInput::MainShoot );
    setButtonValue( NPlayerInput::SecondShoot );


    // Update checkbox
    //bool checked= engine->config ? 1 : 0;
    IGUICheckBox* box = dynamic_cast<IGUICheckBox*>(_overallTab->getElementFromId(NGuiId::FullscreenBox,true));
    box->setChecked( _bufferedParameters.Fullscreen );

// engine().device->isFullscreen()
//


    /////////////////////////////////////////////////////////////////
    ///////     select current resolution
    /////////////////////////////////////////////////////////////////
}



//updateBindDescription()

/**** Function HandleEvents() ****/
bool
COptionState::OnEvent(const irr::SEvent& event)
{

    if( this->waitingNewBind() )
    {

        if( (event.EventType == EET_KEY_INPUT_EVENT) && (event.KeyInput.Key == irr::KEY_ESCAPE))
        {
            // TODO
            //stopWaiting();
            _INFO << "ESCAPE touche, devrait stopper tout bind" ;
        }
//        if( _descriptorGenerator->baseliningBind() )
        else
        {
            //
            //_descriptorGenerator.check();
            if( _descriptorGenerator->processEvent(
                                                      engine().device()->getTimer()->getTime(),
                                                      event
                                                      )
               )
               {
                   // Eat the event
                   return true;
               }
        }
        //! Means this is a unhandled event type

    }


    if( event.EventType == EET_GUI_EVENT)
	{
	    //event.GUIEvent.Caller->getID();
		s32 callerId = event.GUIEvent.Caller->getID();

        switch(event.GUIEvent.EventType){
            case EGET_COMBO_BOX_CHANGED:
                break;

            case EGET_BUTTON_CLICKED:
                _INFO << " id clicked :" << callerId;

                // If we clicked on a bind button, we want to update it
                if(callerId >= NGuiId::Count){

                    if( !waitingNewBind() ){

                        _descriptorGenerator.reset ( new CBindFromEventGenerator(  ) );

                        _buttonToUpdate = static_cast<IGUIButton*>( _tabControl->getElementFromId( callerId, true ) );
                        _buttonToUpdate->setText(L"Waiting...");

                        return true;
                    }
                }

                switch(callerId){

                    case NGuiId::BackToMenu:
                        OnClickBack();
                        return true;

                    case NGuiId::ResetConfig:
                        //engine().ResetConfig();
                        setToConfigValues();
                        break;

                    case NGuiId::SaveConfig:
                        return OnClickSaveButton();
                    //
                    default:

                        break;
                };
                break;


            case EGET_CHECKBOX_CHANGED:

                    switch(callerId){

                        case NGuiId::Antialiasing:
                            break;
                        case NGuiId::VSync:
                            break;
                        case NGuiId::FullscreenBox:
                            break;
                        //case NGuiId::Shaders:
                            break;


                    };

                    break;
            //case :
                        default: break;

            }


        } // endif

//        IGUITab* tab = _tabControl->getTab(_tabControl->getActiveTab());
//        EOptionTab activeTab = static_cast<EOptionTab>(tab->getID());
//        (this->*_processingFunctions[activeTab])(event.GUIEvent);
//
//    if( waitingNewBind() && !_descriptorGenerator->baseliningBind() )
//    {
//        _INFO << "Waiting for new bind";
//
//
//        // generated a valid descriptor event
//        //boost::optional<CBindDescriptor> ret =
//        return _descriptorGenerator->processEvent( engine().device()->getTimer()->getTime(), event);
//
//    }

    return false;
}



/**** Function Update() ****/
void
COptionState::onUpdate(const TTimeUnit& elapsedTime)
{

    // If
    if( waitingNewBind() )
    {
        CBindDescriptor result;

//        _descriptorGenerator->check(
//                                engine().device()->getTimer()->getTime(),
//                                result
//                                    );

        _INFO << "Waiting new bind";

        //! if
        if( _descriptorGenerator->check(engine().device()->getTimer()->getTime(), result) )
        {

            _INFO << "Generation finished";

            std::wstring desc;

            // If already bound
            auto matchingId = std::make_pair( NPlayerInput::MoveLeft,0 );
            auto fullId = UNTRANSLATE_ID( _buttonToUpdate->getID() );

            // If descriptor already used for another bind
//            if( checkForDouble && (*checkForDouble != fullId) ){
            if( GET_MAPPING(1).containBind( result , matchingId) && (matchingId != fullId) )
            {

                result.generateDescription(desc);


                fus::TWCharStringBuilder str("Bind [");

                str( desc.c_str() )(L"] already bound to other bind [");
//                str( BindNames[ checkForDouble->first ] )
                str( BindNames[ matchingId.first ] )
                ("]")
                ;

                //_INFO << "descriptor Bound with id " << BindNames[checkForDouble->first] << " matching input no " << static_cast<int>(checkForDouble->first);
//                _INFO << "fullId " << fullId->first;

                engine().popupMessage(L"/!\\ Bind already in use ", str.str().c_str() );


                GET_MAPPING(1).getBind( fullId ).generateDescription(desc);
                _buttonToUpdate->setText( desc.c_str()  );
            }
            // TODO set new value
            else
            {

                //std::pair<NPlayerInput::EId,int> res = UNTRANSLATE_ID( _buttonToUpdate->getID() );

                GET_MAPPING(1).setBind( fullId, result);
                //description.append( );
                result.generateDescription(desc);
                _buttonToUpdate->setText( desc.c_str() );
            }

            // Stop generating descriptor
            _buttonToUpdate = 0;
            _descriptorGenerator.reset(0);
        }
    }
}

bool
COptionState::processEventForSoundTab(const irr::SEvent::SGUIEvent& event){

    return false;
}


bool
COptionState::processEventForGameplayTab(const irr::SEvent::SGUIEvent& event){
//Language
    return false;
}


bool
COptionState::processEventForGraphicTab(const irr::SEvent::SGUIEvent& event){

    s32 id = event.Caller->getID();
    switch(event.EventType){

        case EGET_CHECKBOX_CHANGED:
            //engine->SwitchBit(NBit::Fullscreen, isChecked());
            //engine->
            break;

        default:
            break;

    };

    return false;
}


std::wstring
COptionState::generateDebugData() const
{
    //
//    stringw desc(L"v");
//    desc += conf::_GAME_VERSION;
//    desc += __DATE__;
//    desc += "\nLuabind version: ";
//    desc += LUABIND_VERSION;
    ICameraSceneNode* cam = scene()->getActiveCamera();
    vector3df camPos = cam->getPosition();

    return std::wstring (L"Resolution");
//    text.append( L"Resolution: ");
    //.append( GameConfig. ).append( L"/" ).append( camPos.Y ).append( L"/" ).append( camPos.Z );

}








/// Create the options menu
void
COptionState::GUI_CreateMainMenu()
{

    IGUIEnvironment* env = gui();


    // parametre le tableau des update
    // TODO utiliser des unary_function ?
    //_processingFunctions[Controls]  = &COptionState::processEventForControlTab;
    _processingFunctions[Graphics]  = &COptionState::processEventForGraphicTab;
    _processingFunctions[Sound]     = &COptionState::processEventForSoundTab;
    _processingFunctions[Gameplay]  = &COptionState::processEventForGameplayTab;


    core::recti rect(0,0,0,0);
    _overallTab = env->addTab( rect, gui()->getRootGUIElement());
    _overallTab->setRelativePositionProportional ( _pos );

    _tabControl = env->addTabControl(rect, _overallTab, true, true);
    _tabControl->setRelativePositionProportional ( core::rectf(0,0,1.f,0.9f));


    //////////////////////////////////////
    //// create control tab
    //////////////////////////////////////

    IGUITab* tab = _tabControl->addTab(L"Controls",Controls);

    createBindLine( NPlayerInput::MoveForward,tab);
    createBindLine( NPlayerInput::MoveBackward,tab);
    createBindLine( NPlayerInput::MoveLeft,tab);
    createBindLine( NPlayerInput::MoveRight,tab);
    createBindLine( NPlayerInput::MainShoot,tab);
    createBindLine( NPlayerInput::SecondShoot,tab);
    createBindLine( NPlayerInput::ModeUp,tab);
    createBindLine( NPlayerInput::ModeDown,tab);

    //////////////////////////////////////
    //// create graphic tab
    //////////////////////////////////////
    createGraphicTab(   _tabControl->addTab(L"Graphics",Graphics)   );
    createSoundTab(     _tabControl->addTab(L"Sound",Sound)         );
    createGameplayTab(  _tabControl->addTab(L"Gameplay",Gameplay)   );

//b->setAlignment( EGUIA_SCALE, EGUIA_SCALE, EGUIA_SCALE, EGUIA_SCALE);

    // create 3 button
    IGUIButton* button = 0;

    recti box( position2di(0,220), dimension2du(100,32) );
              //0,220,100,220 + 32);

    button = env->addButton( box, _overallTab, NGuiId::SaveConfig,
			L"Save", L"Save");

    xDecal(box, 5 );
    //recti(200,220,300,220 + 32)
    button = env->addButton( box, _overallTab, NGuiId::ResetConfig,
			L"Reset", L"Resets configuration");

    xDecal(box, 5 );
    //recti(100,220,200,220 + 32)
    button =  env->addButton( box, _overallTab, NGuiId::BackToMenu,
			L"Back", L"Back to menu");
    //IGUICheckBox* checkBox =  env->addCheckBox();;



    _bufferedParameters = engine().EngineConfig;
    //engine().config->getIrrlichtParameters( _bufferedParameters );
    setToConfigValues();
}


//wchar_t const* name,
void
COptionState::createBindLine(const NPlayerInput::EId& id,IGUITab* tab)
{

        // get specific mapping
//        CBindDescriptor& key1 = engine().GameConfig.Keymap1[id];
//        CBindDescriptor& key2 = engine().GameConfig.Keymap2[id];

        recti pos( 0,id*20,150,(id+1)*20);

        gui()->addStaticText(
                //name,
                BindNames[id],
                //key._name.c_str(),
                pos,    //rect
                true, //border
                false,    //wordwrap
                tab,      // parent
                -1,       // id
                true //fill bg
            );

        xDecal(pos,5);

        //_buttonIds
        //std::pair<int,int> pairIds;
        IGUIButton* button = 0;
        button = gui()->addButton(
                pos
                , tab
                //, 0
                , TRANSLATED_ID(id,0)
                , L""
                , L"Click to change mapping"
                );

        //_buttonsMap.insert( std::make_pair(button,id) );

        xDecal(pos,5);
        //pos.LowerRightCorner
        // TODO afficher le second
        button = gui()->addButton(
                pos
                , tab
                //, 0
                , TRANSLATED_ID(id,1)
                , L"" //GENERATE_DESCRIPTION( Keymap2[id] )
                , L"Click to change mapping"
                );

}


// TODO renommer en Audio
void
COptionState::createSoundTab(IGUITab* tab)
{

    recti box( 10, 45, 200, 60);

    IGUIComboBox* comboBox = gui()->addComboBox(box,tab,NGuiId::Resolution);

    comboBox->addItem(L"Fmodex");
    //comboBox->addItem(L"Null");


    yDecal(box,20);
    IGUIScrollBar* scrollbar = gui()->addScrollBar(true, box , tab, NGuiId::VolumeControl);
    scrollbar->setMax(100);
    // pr le son





    //t3->addChild(scrollbar);
    // TODO choose driver

}



void
COptionState::createGameplayTab(IGUITab* tab)
{


    // checkbox that enables ingame dialogs
    IGUICheckBox* checkbox =  gui()->addCheckBox  (
                0,
                irr::core::recti(10,40,200,70),
                tab,
                NGuiId::SkipDialog,
                L"Skip dialog"
                );


    // list of languages available
    _comboBox = gui()->addComboBox(irr::core::recti( conf::_WINDOW_MARGIN_LEFT,
                                                    conf::_WINDOW_MARGIN_TOP,
                                                    conf::_WINDOW_MARGIN_LEFT + _COMBOBOX_WIDTH,
                                                    _COMBOBOX_HEIGHT
                                                    ),
                                                    tab
                                                    );
    _comboBox->setID( NGuiId::Language );

    //_comboBox->addItem( L"Francais");
    _comboBox->addItem( L"English");

    //_comboBox->setMinSize(core::dimension2du(200,32));

}



////////////////////////////////////////
//////  Graphics
////////////////////////////////////////
void
COptionState::createGraphicTab(IGUITab* tab)
{

    ////////////////////////////////////////
    //////  Resolution tab
    ////////////////////////////////////////

    irr::core::recti box(10,10,200,40);

    _comboBox = gui()->addComboBox(box,tab,NGuiId::Resolution);
    IVideoModeList* list = engine().device()->getVideoModeList();

    core::stringw str;

// TODO selectionner le bon existant
// Ne liste que les modes d'un certain type de profondeur
// += 3
    irr::u32 maxLength = 0;
    //engine->getFont()->getDimension();
    for(s32 i = 0; i < list->getVideoModeCount(); ++i)
    {

        str = videoModeToText(list->getVideoModeResolution(i),list->getVideoModeDepth(i));
        _comboBox->addItem(str.c_str(),i);
        maxLength = std::max(maxLength,str.size());
    }
    //engine->getFont()->getDimension()
    _comboBox->setMinSize(core::dimension2du(maxLength,32));
    //setItemHeight();


    decal(box,0,10);



    ////////////////////////////////////////
    //////  Video driver
    ////////////////////////////////////////
    // TODO
    _comboBox = gui()->addComboBox( box, tab, NGuiId::Resolution );


    std::vector<std::wstring> drivers;
    engine().getAvailableVideoDriverList(drivers);

    BOOST_FOREACH( std::wstring& driverName, drivers )
    {
        //std::cout << p.first << "-->" << p.second << std::endl;
        //if(IIrrlichtDevice::isDriverSupported(EDT_OPENGL)) {
             _comboBox->addItem( driverName.c_str() );//L"OpenGL"
        //}
    }



    //engine->getFont()->getDimension()
    _comboBox->setMinSize(core::dimension2du(maxLength,32));



    decal(box,0,20);


    // ajouter une tickbox pr le fullscreen
    gui()->addCheckBox  ( 0, box, tab, NGuiId::FullscreenBox,
                L"Enable fullscreen"
                );

    decal(box,0,10);

    //IGUICheckBox* checkbox =
    gui()->addCheckBox  (
                0,
                box,
                tab,
                NGuiId::MotionBlurShader,
                L"Enable motion blur shader"
                );

    decal(box,0,10);

    gui()->addCheckBox  (
                0,
                box,
                tab,
                NGuiId::BloomShader,
                L"Enable bloom shader"
                );

    decal(box,0,10);

    gui()->addCheckBox  (
                0,
                box,
                tab,
                NGuiId::VSync,
                L"Enable Vsync"
                );

    decal(box,0,10);

    gui()->addCheckBox  (
                0,
                box,
                tab,
                NGuiId::Antialiasing,
                L"Enable Antialiasing"
                );

}



/**** Function Init() ****/
bool
COptionState::load( CGameEngine* engine,irr::scene::ISceneManager* smgr,irr::video::ITexture* screenshot )
{

    if (IGameState::load( engine,smgr,screenshot ) == false)  return false;

    //_indexOfFocusedElement = -1;
    // TODO setPos
    _pos = conf::OPTIONS_DRAWING_AREA;

//    IBasicGameState::load();

    /// Create initial menu
    GUI_CreateMainMenu();

    MainTab = _overallTab;


    // blurTexture
    video::ITexture* Background = getBackgroundTexture();
    if( Background  != 0)
    {

        // Blur texture
        pp()->blurTexture(Background,Background);

    }


    return true;
}


/**** Function Cleanup() ****/
void
COptionState::unload()
{

    // Effacer tout ce qui a ete charge
    //engine->guienv->get ->removeChild(_tabControl)
    //_tabControl->remove();
//    if(_overallTab) _overallTab->remove();

    IGameState::unload();
};



//modification
void
COptionState::OnClickBack()
{

    //if modification demander a l'user de sauvegarder
    //engine().PopState();
    //remove();
    engine().popState();
}



bool
COptionState::OnClickSaveButton()
{
    //if(!engine->config->RawSave()){
    //! TODO do the same for others parameters or set a global GameConfig struct
    if( !engine().saveConfig( this->GameConfig ) )
    {
        _LOG_ERROR << "TODO: should save config";
    }



    return true;
}




/// \function Draw()
void
COptionState::Draw()
{
//    _INFO << "Draw";

//	driver()->setRenderTarget( ERT_FRAME_BUFFER, true, true );



//    driver()->draw2DImage( texture, pos ,dims);

    engine().getPostProcessing()->render( getBackgroundTexture() );

    _overallTab->draw();

}




irr::core::stringw
COptionState::videoModeToText(const core::dimension2du& dims,const s32& depth)
{


    irr::core::stringw str(dims.Width);// + "x" +dims.Height;
    str += " x ";
    str += dims.Height;
    str += " x ";
    str += depth;
    return str;
}



#undef GET_MAPPING
